Package: devianLM
Type: Package
Title: Detecting Extremal Values in a Normal Linear Model
Version: 1.1.0
Date: 2026-04-21
Authors@R: c(
    person("Guillaume", "Sauliere", 
           role = "aut", 
           email = "guillaumesauliere@hotmail.com",
           comment = c(ORCID = "0000-0001-8263-6456")),
    person("Geoffroy", "Berthelot", 
           role = c("aut", "cre"), 
           email = "geoffroy.berthelot@insep.fr",
           comment = c(ORCID = "0000-0003-4036-6114")),
    person("Jérôme", "Dedecker", 
           role = "aut", 
           email = "jerome.dedecker@u-paris.fr",
           comment = c(ORCID = "0000-0002-8838-0356")))
Description: Provides a method to detect values poorly explained by a Gaussian linear model. The procedure is based on the maximum of the absolute value of the studentized residuals, which is a parameter-free statistic. This approach generalizes several procedures used to detect abnormal values during longitudinal monitoring of biological markers. For methodological details, see: Berthelot G., Saulière G., Dedecker J. (2025). "DEViaN-LM An R Package for Detecting Abnormal Values in the Gaussian Linear Model". HAL Id: hal-05230549. <https://hal.science/hal-05230549>. 
License: GPL-3
Encoding: UTF-8
Imports: Rcpp
LinkingTo: Rcpp, RcppArmadillo
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Depends: R (>= 2.10)
LazyData: true
SystemRequirements: OpenMP (optional, for parallel execution)
NeedsCompilation: yes
Packaged: 2026-04-30 08:50:39 UTC; root
Repository: CRAN
Date/Publication: 2026-04-30 09:40:02 UTC
Author: Guillaume Sauliere [aut] (ORCID:
    <https://orcid.org/0000-0001-8263-6456>),
  Geoffroy Berthelot [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4036-6114>),
  Jérôme Dedecker [aut] (ORCID: <https://orcid.org/0000-0002-8838-0356>)
Maintainer: Geoffroy Berthelot <geoffroy.berthelot@insep.fr>
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 20:19:10 UTC; unix
Archs: devianLM.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,devianlm)
export(devianlm_stats)
export(get_devianlm_threshold)
import(parallel)
importFrom(Rcpp,sourceCpp)
importFrom(stats,.lm.fit)
importFrom(stats,complete.cases)
importFrom(stats,lm.influence)
importFrom(stats,quantile)
importFrom(stats,rnorm)
useDynLib(devianLM, .registration = TRUE)

#' @export
print.devianlm <- function(x, ...) {
  cat("devianLM results\n")
  cat("----------------\n")
  cat("Number of observations:", length(x$reg_residuals), "\n")
  cat("Number of detected outliers:", length(x$outliers), "\n")
  cat("Threshold:", format(x$threshold, digits = 4), "\n")
  cat("\nUse `$` to access full results.\n")
  
  invisible(x)
}

#' Estimate threshold value via Monte-Carlo simulations.
#' 
#' @description
#' Estimates the threshold for the maximum absolute studentized residual in a
#' Gaussian linear model, conditional on the design matrix and 
#' using `n_sims` Monte-Carlo simulation for the quantile of order `quant`.
#' 
#' @details
#' Monte-Carlo simulations are parallelized using OpenMP when available.
#'   
#' @param x either a numeric variable or several numeric variables
#'   (explanatory variables) concatenated in a data frame. 
#'   **Note:** `devianLM` does not add an intercept automatically; 
#'   include a column of ones in `x` if an intercept is desired.
#' @param n_sims Integer. Optional value which is the number of Monte-Carlo simulations. Default is 50,000.
#' @param nthreads Integer. Optional value which is the number of threads to use. Default is
#'   `parallel::detectCores() - 1`.
#' @param quant Numeric. Order of the quantile of interest. Default is 0.95 (this corresponds to a  risk level of 0.05).
#' 
#' @return Numeric value. \item{threshold}{The quantile of order `quant` of the distribution of the maximum of the absolute values of the studentized residuals (depending on the design matrix) is computed via Monte-Carlo simulations (with n_sims simulations).}
#' 
#' @export
#' @useDynLib devianLM, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @import parallel
get_devianlm_threshold <- function( x, n_sims = 50000, nthreads = detectCores() - 1, quant = 0.95 ) {
  sims <- devianlm_cpp(x, n_sims, nthreads)
  quantile(sims, probs = quant)
}

#' Identify outliers using devianLM method
#' 
#' @description 
#' This function determines whether the maximum of the absolute values
#' of the studentized residuals of a Gaussian regression is abnormally high.
#' Outliers are detected by comparing the absolute values of the studentized residuals to a
#' threshold (depending on the design matrix), which can be supplied or estimated
#' via `n_sims` Monte-Carlo simulations.
#' 
#' @param y Numeric. Response vector.
#' @param x either a numeric variable or several numeric variables
#'   (explanatory variables) concatenated in a data frame. `devianLM` does not 
#'   add an intercept automatically; include a column of ones in `x` 
#'   if an intercept is desired.
#' @param threshold Numeric or NULL. If NULL, the threshold value is computed using [get_devianlm_threshold()].
#' @param n_sims Integer. Optional value which is the number of Monte-Carlo simulations. Default is 50,000.
#' @param verbose Logical. If TRUE (default), informative messages are printed during execution (e.g., when ties are detected and handled).
#' @param nthreads Integer. Optional value which is the number of threads to use. Default is
#'   `parallel::detectCores() - 1`.
#' @param quant Numeric. Order of the quantile of interest. Default is 0.95 (this corresponds to a  risk level of 0.05).
#' @param ... Additional arguments passed to [get_devianlm_threshold()].
#' 
#' @details
#' When ties are present in `y`, a small random perturbation is added to avoid
#' numerical issues. The "Ties were detected in the data, 
#' they have been randomly broken" message is displayed when this occurs.
#' 
#' @return devianlm returns an object of class \emph{list} with the following components:
#' \describe{
#'   \item{reg_residuals}{Numeric vector. The studentized residuals from the linear model.}
#'   \item{outliers}{Integer vector. The indices (positions in the original data) of observations identified as outliers based on the threshold.}
#'   \item{threshold}{Numeric value. The cutoff applied to the absolute value of the studentized residuals to flag outliers. If not provided, it is estimated using \code{get_devianlm_threshold()}.}
#'   \item{is_outliers}{Integer vector. A binary vector (0 or 1) of the same length as \code{reg_residuals}, indicating whether each observation is considered an outlier (1) or not (0).}
#' }
#' @importFrom stats .lm.fit lm.influence quantile rnorm complete.cases
#' @import parallel
#' @examples
#' set.seed(123)
#' y <- salary$hourly_earnings_log
#' x <- cbind(1, salary$age, salary$educational_attainment, salary$children_number)
#' 
#' test_salary <- devianlm_stats(y, x, n_sims = 100, quant = 0.95)
#' 
#' plot(test_salary$reg_residuals,
#'   pch = 16, cex = .8,
#'   ylim = c(-1 * max(abs(test_salary$reg_residuals)), max(abs(test_salary$reg_residuals))),
#'   xlab = "", ylab = "Studentized residuals",
#'   col = ifelse(test_salary$is_outliers, "red", "black"))
#' 
#' # Add the thresholds lines:
#' abline(h = c(-test_salary$threshold, test_salary$threshold), col = "chartreuse2", lwd = 2)
#'  
#' @export
devianlm_stats <- function(
    y,
    x,
    threshold = NULL,
    n_sims = 50000,
    verbose = TRUE,
    nthreads = detectCores() - 1,
    quant = 0.95,    
    ... 
) {
  
  if (length(y) != NROW(x)) stop("y and x must have compatible lengths")

  # dealing with NA
  complete_cases <- complete.cases(y, x)
  y_clean <- y[complete_cases]
  x_clean <- x[complete_cases, , drop = FALSE]

  if (is.null(threshold)) {
    threshold <- get_devianlm_threshold( x = x_clean, n_sims = n_sims, nthreads = nthreads, quant = quant )
  }
  
  # add noise if ties are detected
  y_mod <- y_clean
  if (any(duplicated(y_clean))) {
    v <- min(y_clean, na.rm = T) / 1e6
    if (v < 1e-16) v <- .Machine$double.eps
    y_mod <- y_clean + ifelse(duplicated(y_clean), rnorm(length(y_clean), mean = 0, sd = v), 0)
    if (verbose) {
      message("Ties were detected in the data; they have been randomly broken.")
    }
  }
  
  reg <- .lm.fit(x_clean, y_mod)
  reg$qr <- reg
  inf <- lm.influence(reg, do.coef = FALSE)
  reg_residuals <- inf$wt.res / (inf$sigma * sqrt(1 - cbind(inf$hat)))
  
  outliers <- which(abs(reg_residuals) > threshold)
  is_outliers <- integer(length(reg_residuals))
  is_outliers[outliers] <- 1
  
  res <- list(
    reg_residuals = reg_residuals,
    outliers = outliers,
    threshold = threshold,
    is_outliers = is_outliers
  )
  
  class(res) <- "devianlm"
  
  res
  
}

test_that("Package loads correctly", {
  expect_true("devianLM" %in% .packages(all.available = TRUE))
})

test_that("devianlm_stats runs on simple example", {
  set.seed(123)
  x <- as.matrix(rnorm(50))
  y <- 2 * x + rnorm(50)
  
  result <- devianlm_stats(y, x, n_sims = 100) # small n_sims for quick test
  
  expect_true(is.list(result) || is.data.frame(result) || is.numeric(result))
})

#' Salary dataset
#'
#' @description A random sample from the 2012 Current Population Survey (CPS). It is the primary source of labor force statistics for the US population.
#'
#' \itemize{
#'   \item age. age of the individual (0--85)
#'   \item sex. sex of the individual ("F" = Female, "M" = Male)
#'   \item region. region
#'   ("NE" = Northeast, "W" = West, "S" = South, "NW" = Northwest)
#'   \item marital_status. marital status of the individual
#'   ("NM" = Never married, "M" = Married, "D" = Divorced, "S" = Separated, "W" = Widowed)
#'   \item hourly_earnings. how much does the individual earn per hour (00--9999)
#'   \item educational_attainment. educational attainment of the individual
#'   (0 = Children, 31 = Less than 1st grade, 32 = 1st,2nd,3rd,or 4th grade, 33 = 5th or 6th grade, 34 = 7th and 8th grade
#'   35 = 9th grade, 36 = 10th grade, 37 = 11th grade, 38 = 12th grade no diploma, 39 = High school graduate - high school diploma or
#'   equivalent, 40 = Some college but no degree, 41 = Associate degree in college - occupation/vocation program, 42 = Associate degree in college - academic program
#'   43 = Bachelor's degree (for example: BA,AB,BS), 44 = Master's degree (for example: MA,MS,MENG,MED,MSW, MBA), 45 = Professional school degree (for example:MD,DDS,DVM,LLB,JD)
#'   46 = Doctorate degree (for example: PHD,EDD))
#'   \item persons_number. number of persons in household (0--16)
#'   \item children_number. number of children in household (0--9)
#'   \item family_income. family income from basic CPS income screener question
#'   (-1 = Not in universe, 01 = Less than $5,000, 02 = $5,000 to $7,499, 03 = $7,500 to $9,999
#'   04 = $10,000 to $12,499, 05 = $12,500 to $14,999, 06 = $15,000 to $19,999, 07 = $20,000 to $24,999
#'   08 = $25,000 to $29,999, 09 = $30,000 to $34,999, 10 = $35,000 to $39,999, 11 = $40,000 to $49,999
#'   12 = $50,000 to $59,999, 13 = $60,000 to $74,999, 14 = $75,000 to $99,999, 15 = $100,000 to $149,999)
#'   \item hourly_earnings_log. log(hourly_earnings)
#' }
#'
#' @docType data
#' @keywords datasets
#' @name salary
#' @format A data frame with 599 rows and 10 variables
#' @seealso {
#' Original data are available from <https://webapps.ilo.org/surveyLib/index.php/catalog/7379>.
#'
#' The data dictionary is available from <https://www2.census.gov/programs-surveys/cps/datasets/2022/march/asec2022_ddl_pub_full.pdf>.
#' }
#'
"salary"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

devianlm_cpp <- function(X, n_sims, nthreads) {
    .Call(`_devianLM_devianlm_cpp`, X, n_sims, nthreads)
}

#' devianLM: Outlier detection via studentized residuals
#' 
#' Provides a method to detect values poorly explained by a Gaussian linear model. 
#' The procedure is based on the maximum of the absolute value of the studentized 
#' residuals (conditional on the design matrix), which is a parameter-free 
#' statistic. This approach generalizes several procedures used to detect 
#' abnormal values during longitudinal monitoring of biological markers.
#' @details
#' The main R functions are:
#' \itemize{
#'   \item \code{\link{get_devianlm_threshold}}: computes the threshold via Monte-Carlo simulations.
#'   \item \code{\link{devianlm_stats}}: identifies outliers based on this threshold. 
#'   It is also possible to use a custom threshold value, specified by the user.
#' }
#'
#' Monte Carlo simulations are parallelized using OpenMP when available.
#'
#' @references
#' Sauliere, G., Berthelot, G., and Dedecker, J. (2025) \emph{DEViaN-LM An R 
#' Package for Detecting Abnormal Values in the Gaussian Linear Model}. 
#' https://doi.org/10.48550/arXiv.2509.02202
#' \doi{10.48550/arXiv.2509.02202}
#'
#' Berthelot G., Gelein B., Meinadier E. Orhant E., Dedecker J., (2025) 
#' \emph{Z-scores-based methods and their application to 
#' biological monitoring: An extended analysis of professional soccer players 
#' and cyclists athletes}. https://arxiv.org/abs/2510.01810
#' \doi{10.48550/arXiv.2510.01810}
#'
"_PACKAGE"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devianlm_stats.R
\name{get_devianlm_threshold}
\alias{get_devianlm_threshold}
\title{Estimate threshold value via Monte-Carlo simulations.}
\usage{
get_devianlm_threshold(
  x,
  n_sims = 50000,
  nthreads = detectCores() - 1,
  quant = 0.95
)
}
\arguments{
\item{x}{either a numeric variable or several numeric variables
(explanatory variables) concatenated in a data frame.
\strong{Note:} \code{devianLM} does not add an intercept automatically;
include a column of ones in \code{x} if an intercept is desired.}

\item{n_sims}{Integer. Optional value which is the number of Monte-Carlo simulations. Default is 50,000.}

\item{nthreads}{Integer. Optional value which is the number of threads to use. Default is
\code{parallel::detectCores() - 1}.}

\item{quant}{Numeric. Order of the quantile of interest. Default is 0.95 (this corresponds to a  risk level of 0.05).}
}
\value{
Numeric value. \item{threshold}{The quantile of order \code{quant} of the distribution of the maximum of the absolute values of the studentized residuals (depending on the design matrix) is computed via Monte-Carlo simulations (with n_sims simulations).}
}
\description{
Estimates the threshold for the maximum absolute studentized residual in a
Gaussian linear model, conditional on the design matrix and
using \code{n_sims} Monte-Carlo simulation for the quantile of order \code{quant}.
}
\details{
Monte-Carlo simulations are parallelized using OpenMP when available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{salary}
\alias{salary}
\title{Salary dataset}
\format{
A data frame with 599 rows and 10 variables
}
\usage{
salary
}
\description{
A random sample from the 2012 Current Population Survey (CPS). It is the primary source of labor force statistics for the US population.

\itemize{
\item age. age of the individual (0--85)
\item sex. sex of the individual ("F" = Female, "M" = Male)
\item region. region
("NE" = Northeast, "W" = West, "S" = South, "NW" = Northwest)
\item marital_status. marital status of the individual
("NM" = Never married, "M" = Married, "D" = Divorced, "S" = Separated, "W" = Widowed)
\item hourly_earnings. how much does the individual earn per hour (00--9999)
\item educational_attainment. educational attainment of the individual
(0 = Children, 31 = Less than 1st grade, 32 = 1st,2nd,3rd,or 4th grade, 33 = 5th or 6th grade, 34 = 7th and 8th grade
35 = 9th grade, 36 = 10th grade, 37 = 11th grade, 38 = 12th grade no diploma, 39 = High school graduate - high school diploma or
equivalent, 40 = Some college but no degree, 41 = Associate degree in college - occupation/vocation program, 42 = Associate degree in college - academic program
43 = Bachelor's degree (for example: BA,AB,BS), 44 = Master's degree (for example: MA,MS,MENG,MED,MSW, MBA), 45 = Professional school degree (for example:MD,DDS,DVM,LLB,JD)
46 = Doctorate degree (for example: PHD,EDD))
\item persons_number. number of persons in household (0--16)
\item children_number. number of children in household (0--9)
\item family_income. family income from basic CPS income screener question
(-1 = Not in universe, 01 = Less than $5,000, 02 = $5,000 to $7,499, 03 = $7,500 to $9,999
04 = $10,000 to $12,499, 05 = $12,500 to $14,999, 06 = $15,000 to $19,999, 07 = $20,000 to $24,999
08 = $25,000 to $29,999, 09 = $30,000 to $34,999, 10 = $35,000 to $39,999, 11 = $40,000 to $49,999
12 = $50,000 to $59,999, 13 = $60,000 to $74,999, 14 = $75,000 to $99,999, 15 = $100,000 to $149,999)
\item hourly_earnings_log. log(hourly_earnings)
}
}
\seealso{
{
Original data are available from \url{https://webapps.ilo.org/surveyLib/index.php/catalog/7379}.

The data dictionary is available from \url{https://www2.census.gov/programs-surveys/cps/datasets/2022/march/asec2022_ddl_pub_full.pdf}.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devianlm_stats.R
\name{devianlm_stats}
\alias{devianlm_stats}
\title{Identify outliers using devianLM method}
\usage{
devianlm_stats(
  y,
  x,
  threshold = NULL,
  n_sims = 50000,
  verbose = TRUE,
  nthreads = detectCores() - 1,
  quant = 0.95,
  ...
)
}
\arguments{
\item{y}{Numeric. Response vector.}

\item{x}{either a numeric variable or several numeric variables
(explanatory variables) concatenated in a data frame. \code{devianLM} does not
add an intercept automatically; include a column of ones in \code{x}
if an intercept is desired.}

\item{threshold}{Numeric or NULL. If NULL, the threshold value is computed using \code{\link[=get_devianlm_threshold]{get_devianlm_threshold()}}.}

\item{n_sims}{Integer. Optional value which is the number of Monte-Carlo simulations. Default is 50,000.}

\item{verbose}{Logical. If TRUE (default), informative messages are printed during execution (e.g., when ties are detected and handled).}

\item{nthreads}{Integer. Optional value which is the number of threads to use. Default is
\code{parallel::detectCores() - 1}.}

\item{quant}{Numeric. Order of the quantile of interest. Default is 0.95 (this corresponds to a  risk level of 0.05).}

\item{...}{Additional arguments passed to \code{\link[=get_devianlm_threshold]{get_devianlm_threshold()}}.}
}
\value{
devianlm returns an object of class \emph{list} with the following components:
\describe{
\item{reg_residuals}{Numeric vector. The studentized residuals from the linear model.}
\item{outliers}{Integer vector. The indices (positions in the original data) of observations identified as outliers based on the threshold.}
\item{threshold}{Numeric value. The cutoff applied to the absolute value of the studentized residuals to flag outliers. If not provided, it is estimated using \code{get_devianlm_threshold()}.}
\item{is_outliers}{Integer vector. A binary vector (0 or 1) of the same length as \code{reg_residuals}, indicating whether each observation is considered an outlier (1) or not (0).}
}
}
\description{
This function determines whether the maximum of the absolute values
of the studentized residuals of a Gaussian regression is abnormally high.
Outliers are detected by comparing the absolute values of the studentized residuals to a
threshold (depending on the design matrix), which can be supplied or estimated
via \code{n_sims} Monte-Carlo simulations.
}
\details{
When ties are present in \code{y}, a small random perturbation is added to avoid
numerical issues. The "Ties were detected in the data,
they have been randomly broken" message is displayed when this occurs.
}
\examples{
set.seed(123)
y <- salary$hourly_earnings_log
x <- cbind(1, salary$age, salary$educational_attainment, salary$children_number)

test_salary <- devianlm_stats(y, x, n_sims = 100, quant = 0.95)

plot(test_salary$reg_residuals,
  pch = 16, cex = .8,
  ylim = c(-1 * max(abs(test_salary$reg_residuals)), max(abs(test_salary$reg_residuals))),
  xlab = "", ylab = "Studentized residuals",
  col = ifelse(test_salary$is_outliers, "red", "black"))

# Add the thresholds lines:
abline(h = c(-test_salary$threshold, test_salary$threshold), col = "chartreuse2", lwd = 2)
 
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/devianLM-package.R
\docType{package}
\name{devianLM-package}
\alias{devianLM}
\alias{devianLM-package}
\title{devianLM: Outlier detection via studentized residuals}
\description{
Provides a method to detect values poorly explained by a Gaussian linear model.
The procedure is based on the maximum of the absolute value of the studentized
residuals (conditional on the design matrix), which is a parameter-free
statistic. This approach generalizes several procedures used to detect
abnormal values during longitudinal monitoring of biological markers.
}
\details{
The main R functions are:
\itemize{
\item \code{\link{get_devianlm_threshold}}: computes the threshold via Monte-Carlo simulations.
\item \code{\link{devianlm_stats}}: identifies outliers based on this threshold.
It is also possible to use a custom threshold value, specified by the user.
}

Monte Carlo simulations are parallelized using OpenMP when available.
}
\references{
Sauliere, G., Berthelot, G., and Dedecker, J. (2025) \emph{DEViaN-LM An R
Package for Detecting Abnormal Values in the Gaussian Linear Model}.
https://doi.org/10.48550/arXiv.2509.02202
\doi{10.48550/arXiv.2509.02202}

Berthelot G., Gelein B., Meinadier E. Orhant E., Dedecker J., (2025)
\emph{Z-scores-based methods and their application to
biological monitoring: An extended analysis of professional soccer players
and cyclists athletes}. https://arxiv.org/abs/2510.01810
\doi{10.48550/arXiv.2510.01810}
}
\author{
\strong{Maintainer}: Geoffroy Berthelot \email{geoffroy.berthelot@insep.fr} (\href{https://orcid.org/0000-0003-4036-6114}{ORCID})

Authors:
\itemize{
  \item Guillaume Sauliere \email{guillaumesauliere@hotmail.com} (\href{https://orcid.org/0000-0001-8263-6456}{ORCID})
  \item Jérôme Dedecker \email{jerome.dedecker@u-paris.fr} (\href{https://orcid.org/0000-0002-8838-0356}{ORCID})
}

}
